/*
    ES Knot
    Copyright 2005, 2006 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsj
    martin@martin.st
*/

#include <coemain.h>
#include "knotappview.h"
#include <flogger.h>
#include "knot.h"
#include <hal.h>
#include <eikenv.h>
#include <GLES/gl.h>
#include <aknutils.h>
#include <aknnotewrappers.h>
#include <stdlib.h>

void debug(const char *str) {
	TBuf<100> buf;
	while (*str != '\0')
		buf.Append(*str++);
	RFileLogger::WriteFormat(_L("debug"), _L("knotdbg.txt"), EFileLoggingModeAppend, _L("%S"), &buf);
}

CKnotAppView* CKnotAppView::NewL(const TRect& aRect, TBool aFullscreen) {
	CKnotAppView* self = CKnotAppView::NewLC(aRect, aFullscreen);
	CleanupStack::Pop(self);
	return self;
}

CKnotAppView* CKnotAppView::NewLC(const TRect& aRect, TBool aFullscreen) {
	CKnotAppView* self = new(ELeave) CKnotAppView();
	CleanupStack::PushL(self);
	self->ConstructL(aRect, aFullscreen);
	return self;
}

CKnotAppView::CKnotAppView() {
}

CKnotAppView::~CKnotAppView() {
	if (iFont && iGc)
		iGc->DiscardFont();
	delete iBitmap;
	delete iDevice;
	delete iGc;
	delete iDrawer;
	delete iPeriodic;
	delete iKnot;

	eglMakeCurrent(iDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
	eglDestroySurface(iDisplay, iSurface);
	eglDestroyContext(iDisplay, iContext);
	eglTerminate(iDisplay);

	delete [] iFrameTick;

	CloseSTDLIB();
}

void CKnotAppView::ConstructL(const TRect& aRect, TBool aFullscreen) {
	TInt err;

	iDSAStarted = EFalse;
	iEGLInited = EFalse;

	iFrameTickLength = 50;
	iFrameTick = new(ELeave) TInt[iFrameTickLength];
	for (int i = 0; i < iFrameTickLength; i++)
		iFrameTick[i] = 0;

	iDrawMode = EEGLWindow;

	CreateWindowL();
	iOrigRect = aRect;
	if (aFullscreen) {
		SetExtentToWholeScreen();
		iWinMode = EFullscreen;
	} else {
		SetRect(aRect);
		iWinMode = EWindowed;
	}
	ActivateL();

//	SetupBitmap();


	if (HAL::Get(HALData::ESystemTickPeriod, iPeriod) != KErrNone)
		iPeriod = 1000/64;
	else
		iPeriod /= 1000;


	iKnot = new Knot(Rect().Width(), Rect().Height());
	iKnot->setUpdateVertices(true);
	iKnot->setTexMode(Knot::TEXSPEC);

	iPeriodic = CPeriodic::NewL(CActive::EPriorityIdle);
	iStarted = EFalse;
	SetForeground(ETrue);

	iFrame = 0;
	iLastFPSFrame = 0;
	iLastFPSTime = (TUint)-1;
	iTime = 0;
	iLastFPS = 0;
	iLastResetTick = 0;
	iLastAfterTick = 0;

	RFs &fs = iEikonEnv->FsSession();
	RFile file;
	_LIT(KFileName, "c:\\logs\\debug\\gles.txt");
	TInt flags = EFileWrite | EFileStreamText;
	err = file.Replace(fs, KFileName, flags);
	if (err == KErrNone) {
		const GLubyte* str = glGetString(GL_VENDOR);
		TBuf8<500> buf;
		buf.Format(_L8("vendor: %s\n"), str);
		file.Write(buf);

		str = glGetString(GL_RENDERER);
		buf.Format(_L8("renderer: %s\n"), str);
		file.Write(buf);

		str = glGetString(GL_VERSION);
		buf.Format(_L8("version: %s\n"), str);
		file.Write(buf);

		str = glGetString(GL_EXTENSIONS);
		buf.Format(_L8("extensions: %s\n"), str);
		file.Write(buf);
		file.Close();
	}

}

static TBool Better(TInt aSize, TInt aRed, TInt aAlpha, TInt aBestSize, TInt aBestRed, TInt aBestAlpha) {
	// sort according to total size, red size, alpha
	if (aSize > aBestSize) return EFalse;
	if (aSize < aBestSize) return ETrue;
	if (aRed > aBestRed) return EFalse;
	if (aRed < aBestRed) return ETrue;
	if (aAlpha > aBestAlpha) return EFalse;
	if (aAlpha < aBestAlpha) return ETrue;
	return EFalse;
}

void CKnotAppView::InitEGL(TBool aPixmap) {
	EGLConfig config;
	TBool reinit = EFalse;

	if (!iEGLInited) {
		iDisplay = eglGetDisplay(EGL_DEFAULT_DISPLAY);
		if (!iDisplay) {
			_LIT(KGetDisplayFailed, "eglGetDisplay failed");
			User::Panic(KGetDisplayFailed, eglGetError());
		}

		if (eglInitialize(iDisplay, NULL, NULL) == EGL_FALSE) {
			_LIT(KInitializeFailed, "eglInitialize failed");
			User::Panic(KInitializeFailed, eglGetError());
		}

		iEGLInited = ETrue;
	} else {
		eglMakeCurrent(iDisplay, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
		eglDestroySurface(iDisplay, iSurface);
		eglDestroyContext(iDisplay, iContext);
		reinit = ETrue;
	}

	EGLConfig *configList = NULL;
	EGLint numConfigs = 0, configSize = 0;

	if (eglGetConfigs(iDisplay, configList, configSize, &numConfigs) == EGL_FALSE) {
		_LIT(KGetConfigsFailed, "eglGetConfigs failed");
		User::Panic(KGetConfigsFailed, eglGetError());
	}

	configSize = numConfigs;
	configList = (EGLConfig*) User::Alloc(sizeof(EGLConfig)*configSize);
	if (!configList) {
		_LIT(KConfigAllocFailed, "config alloc failed");
		User::Panic(KConfigAllocFailed, eglGetError());
	}

	TDisplayMode mode = Window().DisplayMode();
	TInt bufferSize = 0;
	TInt minRed = 0;

	switch (mode) {
		case EColor4K: bufferSize = 12; minRed = 4; break;
		case EColor64K: bufferSize = 16; minRed = 5; break;
		case EColor16M: bufferSize = 24; minRed = 8; break;
		case EColor16MU: bufferSize = 32; minRed = 8; break;
		default:
			_LIT(KModeError, "unsupported displaymode");
			User::Panic(KModeError, 0);
			break;
	}
	RFileLogger::WriteFormat(_L("debug"), _L("egl.txt"), EFileLoggingModeAppend, _L("mode: %d bs: %d mr: %d"), mode, bufferSize, minRed);

	EGLint type = EGL_WINDOW_BIT;
	if (aPixmap)
		type = EGL_PIXMAP_BIT;
	EGLint attribList[] = {
		EGL_BUFFER_SIZE, bufferSize, EGL_DEPTH_SIZE, 16,
		EGL_SURFACE_TYPE, type,
		EGL_NONE
	};

	if (eglChooseConfig(iDisplay, attribList, configList, configSize, &numConfigs) == EGL_FALSE) {
		_LIT(KChooseConfigFailed, "eglChooseConfig failed");
		User::Panic(KChooseConfigFailed, eglGetError());
	}

	int best = -1;
	int bestRed = 0;
	int bestSize = 0;
	int bestAlpha = 0;
	for (int n = 0; n < numConfigs; n++) {
		EGLint red, green, blue, alpha, size;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_RED_SIZE, &red) != EGL_TRUE) continue;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_GREEN_SIZE, &green) != EGL_TRUE) continue;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_BLUE_SIZE, &blue) != EGL_TRUE) continue;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_ALPHA_SIZE, &alpha) != EGL_TRUE) continue;
		if (eglGetConfigAttrib(iDisplay, configList[n], EGL_BUFFER_SIZE, &size) != EGL_TRUE) continue;


		if (red < minRed) continue;

		RFileLogger::WriteFormat(_L("debug"), _L("egl.txt"), EFileLoggingModeAppend, _L("rgbas: %d %d %d %d %d - %d"), red, green, blue, alpha, size, n);

		if (best < 0) {
			best = n;
			bestRed = red;
			bestSize = size;
			bestAlpha = alpha;
		} else {
			if (Better(size, red, alpha, bestSize, bestRed, bestAlpha)) {
//			if (size <= bestSize && red <= bestRed && alpha <= bestAlpha) {
				best = n;
				bestRed = red;
				bestSize = size;
				bestAlpha = alpha;
			}
		}
	}
	if (best < 0) {
		_LIT(KNoSuitableConfigs, "No suitable configs");
		User::Panic(KNoSuitableConfigs, 0);
	}
	RFileLogger::WriteFormat(_L("debug"), _L("egl.txt"), EFileLoggingModeAppend, _L("using config %d"), best);

	config = configList[best];
	User::Free(configList);

	iContext = eglCreateContext(iDisplay, config, NULL, NULL);
	if (!iContext) {
		_LIT(KCreateContextFailed, "eglCreateContext failed");
		User::Panic(KCreateContextFailed, eglGetError());
	}

	if (aPixmap)
		iSurface = eglCreatePixmapSurface(iDisplay, config, iBitmap, NULL);
	else
		iSurface = eglCreateWindowSurface(iDisplay, config, &Window(), NULL);

	if (!iSurface) {
		_LIT(KCreateWindowSurfaceFailed, "eglCreateWindowSurface failed");
		_LIT(KCreatePixmapSurfaceFailed, "eglCreatePixmapSurface failed");
		if (aPixmap)
			User::Panic(KCreatePixmapSurfaceFailed, eglGetError());
		else
			User::Panic(KCreateWindowSurfaceFailed, eglGetError());
	}

	if (eglMakeCurrent(iDisplay, iSurface, iSurface, iContext) == EGL_FALSE) {
		_LIT(KMakeCurrentFailed, "eglMakeCurrent failed");
		User::Panic(KMakeCurrentFailed, eglGetError());
	}

	if (reinit && iKnot)
		iKnot->reinit();
}

void CKnotAppView::HandleResourceChange(TInt aType) {
	CCoeControl::HandleResourceChange(aType);
	if (aType == KEikDynamicLayoutVariantSwitch) {
		iOrigRect = static_cast<CEikAppUi*>(iEikonEnv->AppUi())->ClientRect();
		if (iWinMode == EFullscreen)
			SetExtentToWholeScreen();
		else
			SetRect(iOrigRect);
	}
}

void CKnotAppView::SetWindowMode(TWindowMode aMode) {
	if (aMode == iWinMode)
		return;

	iWinMode = aMode;

	if (iWinMode == EFullscreen)
		SetExtentToWholeScreen();
	else
		SetRect(iOrigRect);

}

void CKnotAppView::SizeChanged() {
	if (iKnot)
		iKnot->setSize(Rect().Width(), Rect().Height());

	SetupBitmap();
}

void CKnotAppView::SetupBitmap() {
	if (iFont && iGc)
		iGc->DiscardFont();
	delete iBitmap;
	delete iDevice;
	delete iGc;

	iBitmap = new(ELeave) CWsBitmap(iEikonEnv->WsSession());
	iBitmap->Create(Rect().Size(), iEikonEnv->ScreenDevice()->DisplayMode());

	iDevice = CFbsBitmapDevice::NewL(iBitmap);
	TInt err = iDevice->CreateContext(iGc);
	if (err != KErrNone) {
		_LIT(KPanicMsg, "CFbsBitmapDevice::CreateContext");
		User::Panic(KPanicMsg, err);
	}

//	iFont = LatinBold13();
	iFont = LatinPlain12();
//	iFont = NumberPlain5();
	iGc->UseFont(iFont);

	SetDrawMode(iDrawMode, ETrue);
}

void CKnotAppView::SetDrawMode(TDrawMode aMode, TBool aReset) {
	if (aMode == iDrawMode && !aReset)
		return;

	delete iDrawer;
	iDrawer = NULL;

	TDrawMode lastMode = iDrawMode;
	iDrawMode = aMode;

	TInt err;
	switch (iDrawMode) {
	case EEGLWindow:
		InitEGL(EFalse);
		break;

	case EDoubleBuf:
		if (lastMode == EEGLWindow || aReset)
			InitEGL(ETrue);
		break;

	case EDSA:
		if (lastMode == EEGLWindow || aReset)
			InitEGL(ETrue);
		iDrawer = CDirectScreenAccess::NewL(iEikonEnv->WsSession(), *(iEikonEnv->ScreenDevice()), Window(), *this);
		iEikonEnv->WsSession().Flush();
		TRAP(err, iDrawer->StartL());
		if (err == KErrNone) {
			iDrawer->ScreenDevice()->SetAutoUpdate(ETrue);
			iDSAStarted = ETrue;
		}
		break;
	}


}


void CKnotAppView::SetForeground(TBool aForeground) {
	iForeground = aForeground;
	SetDrawing(aForeground);
}

void CKnotAppView::SetDrawing(TBool aDrawing) {
	if (!iForeground)
		aDrawing = EFalse;
	if (aDrawing) {
		if (!iStarted) {
			iPeriodic->Start(100, 100, TCallBack(CKnotAppView::StaticCallback, this));
			iLastTick = User::TickCount();
		}
	} else {
		if (iStarted)
			iPeriodic->Cancel();
	}
	iStarted = aDrawing;
}

void CKnotAppView::Draw(const TRect& aRect) const {
	CWindowGc& gc = SystemGc();
//	TRect rect = Rect();
//	gc.Clear(rect);
	if (iDrawMode == EEGLWindow) {
		eglCopyBuffers(iDisplay, iSurface, iBitmap);
		eglWaitGL();
	}
	gc.BitBlt(TPoint(0, 0), iBitmap);
}

void CKnotAppView::Update() {
	TInt err = KErrNone;
	if (!iDSAStarted) {
		TRAP(err, iDrawer->StartL());
		if (err == KErrNone) {
			iDrawer->ScreenDevice()->SetAutoUpdate(ETrue);
			iDSAStarted = ETrue;
		}
	}
	if (err == KErrNone) {
		iDrawer->Gc()->BitBlt(TPoint(0, 0), iBitmap);
	}
}

void CKnotAppView::Restart(RDirectScreenAccess::TTerminationReasons aReason) {
	SetDrawing(ETrue);
}

void CKnotAppView::AbortNow(RDirectScreenAccess::TTerminationReasons aReason) {
	SetDrawing(EFalse);
	iDSAStarted = EFalse;
}


TInt CKnotAppView::StaticCallback(TAny* aArg) {
	CKnotAppView* view = (CKnotAppView*) aArg;
	view->Callback();
	return 0;
}

void CKnotAppView::Callback() {
	iFrame++;

	iKnot->draw(iTime);

	TInt tick = User::TickCount();
	iRelTick += tick - iLastTick;
	iTime = iRelTick*iPeriod;
	iLastTick = tick;

	if (iDrawMode == EDoubleBuf || iDrawMode == EDSA) {
		eglWaitGL();
		TPoint point(2, 2 + iFont->HeightInPixels());
		TBuf<20> buf;
		buf.Format(_L("%d fps"), iLastFPS);
		iGc->SetPenColor(KRgbWhite);
		iGc->SetDrawMode(CGraphicsContext::EDrawModeXOR);
		iGc->DrawText(buf, point);
		TInt ticklength = iRelTick - iFrameTick[iFrame % iFrameTickLength];
		buf.Format(_L("%d ticks/%d frames"), ticklength, iFrameTickLength);
		point.iX = iBitmap->SizeInPixels().iWidth - 2 - iFont->TextWidthInPixels(buf);
		iGc->DrawText(buf, point);
		eglWaitNative(EGL_CORE_NATIVE_ENGINE);
	}
	iFrameTick[iFrame % iFrameTickLength] = iRelTick;

//	eglCopyBuffers(iDisplay, iSurface, iBitmap);
	if (iDrawMode == EEGLWindow)
		eglSwapBuffers(iDisplay, iSurface);
	else if (iDrawMode == EDoubleBuf) {
		DrawNow();
	} else if (iDrawMode == EDSA) {
		Update();
	}


	if (iLastFPSTime == (TUint)-1)
		iLastFPSTime = iTime;

	if ((tick - iLastResetTick)*iPeriod >= 3000) {
		User::ResetInactivityTime();
		iLastResetTick = tick;
	}
	if ((tick - iLastAfterTick)*iPeriod >= 1500) {
		User::After(0);
		iLastAfterTick = tick;
	}

	if (iTime - iLastFPSTime > 1000) {
		iLastFPS = 1000*(iFrame - iLastFPSFrame)/(iTime - iLastFPSTime);
		RFileLogger::WriteFormat(_L("debug"), _L("knot.txt"), EFileLoggingModeAppend, _L("%d fps"), iLastFPS);
		iLastFPSTime = iTime;
		iLastFPSFrame = iFrame;
	}
}

TKeyResponse CKnotAppView::OfferKeyEventL(const TKeyEvent& aKeyEvent, TEventCode aType) {
	if (aType != EEventKeyDown)
		return EKeyWasNotConsumed;

	if (aKeyEvent.iScanCode == EStdKeyDevice3) {
		if (iKnot)
			iKnot->startMorph();
		return EKeyWasConsumed;
	}
	return EKeyWasNotConsumed;
}

